use defmt::*;
use embassy_futures::select::*;
use embassy_rp::gpio;
use embassy_rp::pac;
use embassy_rp::interrupt::UART0_IRQ;
use embassy_time::{Duration, Timer};
use embassy_embedded_hal::SetConfig;
use embassy_rp::peripherals::{self, UART0};
use embassy_rp::uart::{self, UartRx};
use embassy_rp::pio::{Config, Pio, ShiftConfig, ShiftDirection, FifoJoin};
use embassy_rp::relocate::RelocatedProgram;
use embassy_rp::Peripheral;
use {defmt_rtt as _, panic_probe as _};
use fixed::traits::ToFixed;
use fixed_macro::types::U56F8;

use crate::dont_abort::DontAbort;

pub struct RS485 {
    pio: peripherals::PIO0,
    rx_pin: peripherals::PIN_17,
    dma_channel: peripherals::DMA_CH0,
    rx: UartRx<'static, UART0, uart::Async>
}

fn pin_io<P: gpio::Pin>(pin: &P) -> pac::io::Gpio {
    use gpio::Bank;

    let block = match pin.bank() {
        Bank::Bank0 => pac::IO_BANK0,
        Bank::Qspi => pac::IO_QSPI,
    };
    block.gpio(pin.pin() as _)
}

async fn debug_print_pio_addr(sm: pac::pio::StateMachine) {
    let mut prev = 42u8;
    loop {
        let addr = unsafe { sm.addr().read().addr() };
        if prev != addr {
            prev = addr;
            info!("SM addr: {}", addr);
        }
        Timer::after(Duration::from_millis(200)).await;
    }
}

impl RS485 {
    pub fn new(
            uart: UART0, rx: peripherals::PIN_17, rx_irq: UART0_IRQ, rx_dma_channel: peripherals::DMA_CH1, uart_config: uart::Config,
            pio: peripherals::PIO0, dma_channel: peripherals::DMA_CH0
        ) -> RS485 {
        // SAFETY: The auto-baud will only read from this pin and we will set it back to UART mode after initialising the PIO.
        let rx_pin_for_autobaud = unsafe { rx.clone_unchecked() };

        let uart_rx = UartRx::new(
            uart,
            rx,
            rx_irq,
            rx_dma_channel,
            uart_config,
        );
        RS485 { pio, rx_pin: rx_pin_for_autobaud, dma_channel, rx: uart_rx }
    }

    pub async fn run_task(mut self: Self) {
        let Pio {
            mut common,
            sm0: mut sm,
            mut irq0,
            ..
        } = Pio::new(self.pio);

        sm.set_enable(false);

        let prg_set_osr = pio_proc::pio_asm!(
            ".origin 0",
            "pull",
            "hang:",
            "jmp hang",
        );

        let prg = pio_proc::pio_asm!(
            ".origin 0",

            // Send zero to RX FIFO to signal start of next measurement.
            ".wrap_target",
            "timeout:",
            "in null, 32",

            // Wait for pin to be high for a longer time (UART is idle), copy waiting time to X
            // and decrement until done, start again when pin changes.
            "wait_idle:",
            "wait 1 pin 0",
            "mov x, osr",
            "continue_wait_idle:",
            "jmp pin still_one",
            "jmp wait_idle",  // we got a zero, start again
            "still_one:",
            "jmp x--, continue_wait_idle",

            // We have a long 1, i.e. UART is idle. Wait for start bit.
            "have_long_high:",
            "wait 0 pin 0",
            "nop [4]",
            "jmp pin wait_idle",  // abort if zero was just a glitch

            // We want to measure 9 bits (start bit plus eight (or nine) data bits). However, we need
            // different code for high and low so it's easier to measure 4x2 bits. Weuse Y as the loop counter.
            "set y, 3",

            // Start counting. We count down from the value in OSR (which is our timeout).
            "measure_next:",
            "mov x, osr",
            "measure_low:",
            "jmp pin changed_to_one",
            //"nop",
            "jmp x-- measure_low",
            "jmp timeout",
            "changed_to_one:",
            "in x, 32"

            // Start counting, again, but for a high pulse.
            "mov x, osr",
            "measure_high:",
            "jmp pin still_high2",
            "jmp changed_to_low",
            "still_high2:",
            "jmp x-- measure_high",
            "jmp timeout",
            "changed_to_low:",
            "in x, 32",
            "jmp y-- measure_next",

            // We are done.
            "irq set 0",
            ".wrap",
        );

        let relocated = RelocatedProgram::new(&prg_set_osr.program);
        let mut cfg = Config::default();
        let loaded_program = common.load_program(&relocated);
        cfg.use_program(&loaded_program, &[]);
        const SM_FREQ: u32 = 125_000_000;
        cfg.clock_divider = (U56F8!(125_000_000) / U56F8!(125_000_000 /* SM_FREQ */)).to_fixed();
        cfg.shift_in = ShiftConfig {
            auto_fill: true,
            threshold: 32,
            direction: ShiftDirection::Left,
        };
        cfg.shift_out = ShiftConfig {
            auto_fill: false,
            threshold: 32,
            direction: ShiftDirection::Right,
        };
        let rx_pin_pio = common.make_pio_pin(unsafe { self.rx_pin.clone_unchecked() });
        cfg.set_in_pins(&[&rx_pin_pio]);
        cfg.set_jmp_pin(&rx_pin_pio);

        sm.set_config(&cfg);
        sm.set_enable(true);

        // Timeout: Modbus wants 1.5 ms between frames so we make this a bit smaller. SM runs at 25 MHz
        // but we need two clocks per loop.
        let timeout_start_value = (SM_FREQ as f32 * 1.2e-3 / 2.) as u32;
        sm.tx().push(timeout_start_value);
        info!("timeout_start_value: {} = 0x{:08x}", timeout_start_value, timeout_start_value);

        // switch to the real program and join FIFOs
        unsafe { common.free_instr(loaded_program.used_memory); };
        sm.set_enable(false);
        let relocated = RelocatedProgram::new(&prg.program);
        let loaded_program = common.load_program(&relocated);
        cfg.use_program(&loaded_program, &[]);
        cfg.fifo_join = FifoJoin::RxOnly;
        sm.set_config(&cfg);
        sm.set_enable(true);

        // set rx pin function back to UART
        // (PIO can always read and we don't want to write but embassy_rp cannot know that so it "helpfully" configured the pin for PIO function.)
        unsafe {
            pin_io(&self.rx_pin).ctrl().write(|w| w.set_funcsel(embassy_rp::pac::io::vals::Gpio17ctrlFuncsel::UART0_RX.0));
        }

        info!("Program size for auto-baud: {}", prg.program.code.len());

        let mut dma_in_ref = self.dma_channel.into_ref();
        let mut din = [42u32; 9];
        let mut bit_index = 0;
        let mut rx_buf = [0; 1];
        let mut rx_future = DontAbort::new(self.rx.read(&mut rx_buf));
        loop {
            let x = select4(
                irq0.wait(),
                sm.rx().dma_pull(dma_in_ref.reborrow(), &mut din),
                debug_print_pio_addr(embassy_rp::pac::PIO0.sm(0)),
                rx_future.continue_wait(),
            ).await;
            match x {
                Either4::First(_) => {
                    let mut sum = 0;
                    let mut first = 0;
                    let mut ok = false;
                    let mut got_to_eight = false;
                    for i in 0..din.len() {
                        if din[i] == 0 {
                            bit_index = 0;
                            info!("SM in {}: {:08x} -> start", i, din[i]);
                        } else {
                            let mut delay = timeout_start_value - din[i];
                            if bit_index == 0 {
                                delay += 7;
                            } else if bit_index%2 == 1 {
                                //delay += 4;
                                delay += 3;
                            } else {
                                delay += 3;
                            }
                            let millis = (delay) as f32 / SM_FREQ as f32 * 1000. * 2.;
                            let baud = 1000. / millis;
                            info!("SM in {} ({}): {:08x} -> {} -> {} ms -> {}", i, bit_index, din[i], delay, millis, baud);
                            if bit_index == 0 {
                                sum = 0;
                                first = delay;
                                ok = true;
                            } else if delay < first-first/4 || delay > first+first/4 {
                                ok = false;
                            }
                            sum += delay;
                            bit_index += 1;
                            got_to_eight = bit_index == 8;
                        }
                    }

                    for _ in bit_index..8 {
                        if let Some(x) = sm.rx().try_pull() {
                            let mut delay = timeout_start_value - x;
                            if bit_index == 0 {
                                delay += 7;
                            } else if bit_index%2 == 1 {
                                //delay += 4;
                                delay += 3;
                            } else {
                                delay += 3;
                            }
                            let millis = (delay) as f32 / SM_FREQ as f32 * 1000. * 2.;
                            let baud = 1000. / millis;
                            info!("SM in ({}): {:08x} -> {} -> {} ms -> {}", bit_index, x, delay, millis, baud);
                            if bit_index == 0 {
                                sum = 0;
                                first = delay;
                                ok = true;
                            } else if delay < first-first/4 || delay > first+first/4 {
                                ok = false;
                            }
                            sum += delay;
                            bit_index += 1;
                            got_to_eight = bit_index == 8;
                        } else {
                            break
                        }
                    }

                    if got_to_eight && ok {
                        let millis = sum as f32 / SM_FREQ as f32 * 1000. * 2. / 8.;
                        let baud = 1000. / millis;
                        info!("Guessed baud rate: {}", baud);
                    }
                },
                Either4::Second(_) => {
                    if false {
                        for i in 0..din.len() {
                            if din[i] == 0 {
                                bit_index = 0
                            } else {
                                bit_index += 1
                            }
                            info!("SM in {}: {:08x}", i, din[i]);
                        }
                    }
                },
                Either4::Fourth(x) => {
                    drop(rx_future);
                    match x {
                        Result::Ok(()) => {
                            info!("RX {:?}", rx_buf);
                        },
                        Result::Err(e) => {
                            info!("RX error {:?}", e);
                        },
                    }
                    rx_future = DontAbort::new(self.rx.read(&mut rx_buf));
                },
                _ => {
                },
            }
        }
    }
}
